/*
 *   Copyright (C) 2001  mohamed Elassadi <mohamed@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gnome.h>
#include "gm-entry.h"

#define DEFAULT_HISTORY_FILE_NAME ".gmyclient-history"
#define DEFAULT_COMPLETION_FILE_NAME ".gmyclient-completion"
#define MAXLINES 1000
#define REMOVE_MESSAGE_TIMEOUT 700


enum {
    MOVE_UP =0,
    MOVE_DOWN,
    MOVE_LAST,
};

static void    gm_entry_class_init         (GmEntryClass *klass);
static void    gm_entry_init               (GmEntry      *entry);

static gint    gm_entry_key_press          (GtkWidget    *widget, GdkEventKey    *event);

static void    gm_entry_size_allocate      (GtkWidget     *widget, GtkAllocation  *allocation);
static void    gm_entry_size_request       (GtkWidget     *widget, GtkRequisition *requisition);

static GList * entry_load_completion       (void);

static void    gm_entry_destroy            (GtkObject *object);
static void    entry_dynamic_expansion     (GmEntry   *entry);
static gchar * entry_get_completion_prefix (GtkEditable *editable);
static void    entry_cycle_expand          (GmEntry *entry);
static gint    entry_hide_message (gpointer data) ;
static void    entry_show_message (GmEntry * entry, gchar * message) ;
static void    entry_insert_text         (GtkEditable       *editable,
        const gchar       *new_text,
        gint               new_text_length,
        gint              *position);


static GtkWidgetClass *parent_class = NULL;

void (* parent_insert_text)  (GtkEditable    *editable,
                              const gchar    *text,
                              gint            length,
                              gint           *position);

guint
gm_entry_get_type (void)
{
    static guint entry_type = 0;

    if (!entry_type)
    {
        static const GtkTypeInfo entry_info =
            {
                "GmEntry",
                sizeof (GmEntry),
                sizeof (GmEntryClass),
                (GtkClassInitFunc) gm_entry_class_init,
                (GtkObjectInitFunc) gm_entry_init,
                /* reserved_1 */ NULL,
                /* reserved_2 */ NULL,
                (GtkClassInitFunc) NULL,
            };

        entry_type = gtk_type_unique (gtk_entry_get_type (), &entry_info);
    }

    return entry_type;
}



static void
gm_entry_class_init (GmEntryClass *class)
{
    GtkWidgetClass *widget_class;
    GtkObjectClass *object_class;
    GtkEditableClass *editable_class;
    widget_class = (GtkWidgetClass*) class;
    parent_class = gtk_type_class (gtk_entry_get_type ());
    object_class = (GtkObjectClass *) class;
    editable_class = (GtkEditableClass*) class;

    widget_class->key_press_event = gm_entry_key_press;
    widget_class->size_request    = gm_entry_size_request;
    widget_class->size_allocate   = gm_entry_size_allocate;
    parent_insert_text = editable_class->insert_text;
    editable_class->insert_text = entry_insert_text;
    object_class->destroy = gm_entry_destroy;
}


static void
gm_entry_init (GmEntry *entry){


    entry->completion_list = entry_load_completion();
    entry->completion      = g_completion_new (NULL);
    if (entry->completion_list)
        g_completion_add_items (entry->completion, entry->completion_list);
    entry->compl_cycling   = FALSE;
    entry->remember_message = NULL;
    entry->remember_cursor_pos = 0;
    entry->remove_messages_timer_tag = 0;

}

GtkWidget*
gm_entry_new (void)
{
    GmEntry *entry;

    entry = gtk_type_new (gm_entry_get_type ());
    return GTK_WIDGET (entry);
}





static gint gm_entry_key_press (GtkWidget   *widget,
                                GdkEventKey *event)
{
    GmEntry * entry ;
    GtkEditable *editable;


    entry =  GM_ENTRY(widget);
    editable = GTK_EDITABLE (widget);

    if (entry->remove_messages_timer_tag )
    {
        entry_hide_message (entry);
        return TRUE;
    }

    if ( (event->keyval == GDK_Left))
    {
        gint  cursor_pos = editable->current_pos;
        if (cursor_pos <= 0)
        {
            entry_show_message (entry, "Beginning of buffer");
            return TRUE;
        }
    }

    if ( (event->keyval == GDK_Right))
    {
        gint  cursor_pos = editable->current_pos;
        gchar * text ;
        text = gtk_entry_get_text (GTK_ENTRY (widget));
        if (!text || (strlen(text) == cursor_pos))
        {
            entry_show_message (entry, "End of buffer");
            return TRUE;
        }
    }


    if (event->state & GDK_CONTROL_MASK)
    {
        switch (event->keyval)
        {
        case GDK_y :
            {		/* paste the ringdata */
                gtk_selection_convert(widget, GDK_SELECTION_PRIMARY,
                                      GDK_TARGET_STRING, 0);
                return TRUE;
                break;
            }
        case GDK_c :
            {		/* clear the line */
                gtk_editable_delete_text (GTK_EDITABLE (widget), 0, -1);
                return TRUE;
                break;
            }
        }
    }






    if ((* GTK_WIDGET_CLASS (parent_class)->key_press_event) (widget, event))
    {
        GM_ENTRY (widget)->compl_cycling = FALSE;
        return TRUE;
    }


    if ((event->keyval == GDK_Tab) )
    {
        entry_dynamic_expansion (GM_ENTRY (widget));
        return TRUE;		/* return TRUE to stop event-propag */
    }
    GM_ENTRY (widget)->compl_cycling = FALSE;




    return FALSE;
}




static void
gm_entry_size_allocate (GtkWidget     *widget,
                        GtkAllocation *allocation)
{

    if (GTK_WIDGET_CLASS (parent_class)->size_allocate)
        (* GTK_WIDGET_CLASS (parent_class)->size_allocate) (widget, allocation);
}


static void
gm_entry_size_request (GtkWidget      *widget,
                       GtkRequisition *requisition)
{

    if (GTK_WIDGET_CLASS (parent_class)->size_request)
        (* GTK_WIDGET_CLASS (parent_class)->size_request) (widget, requisition);

}



static GList * entry_load_file_as_list (gchar *filename)
{

    gchar *line   = NULL;
    gchar *buffer = NULL;
    FILE  *file   = NULL;
    GList * list = NULL;
    gint   i =0;
    if ((file = fopen(filename,"r"))!=NULL)
    {
        buffer = (gchar *) g_malloc(1024);
        while (fgets(buffer,1023, file) && i<= MAXLINES )
        {
            if (!buffer) break;
            if (*buffer  !='\n' && *buffer != '#')
            {
                g_strdelimit(buffer, "\n",'\0');
                line = g_strdup(buffer);
                list = g_list_append(list,line);
                i++;
            }
        }
        fclose(file);
    }
    g_free(buffer);

    return list;
}
static GList * entry_load_completion (void)
{


    gchar * home_dir;
    gchar * filename;
    GList * list = NULL;

    home_dir = g_get_home_dir();
    filename = g_strdup_printf("%s/%s",home_dir,DEFAULT_COMPLETION_FILE_NAME);
    list = entry_load_file_as_list(filename);
    g_free (filename);
    return list;
}


static void
gm_entry_destroy (GtkObject * object)
{

    GmEntry * entry;



    entry = GM_ENTRY (object);
    if (GTK_OBJECT_CLASS (parent_class)->destroy)
        (*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}




/* -------------------------------------------------------------------------------- Completion stuff ----------------------------------------*/


/* this function return a new allocated prefix must be freed by the caller */

static gchar * entry_get_completion_prefix (GtkEditable *editable)
{

    gint cursor_pos=0;
    gchar * text=NULL;
    gchar *s;
    gchar *prefix=NULL;
    gint i=0;

    g_return_val_if_fail (editable != NULL, NULL);

    cursor_pos = editable->current_pos;
    if (!cursor_pos) return NULL;

    text = gtk_editable_get_chars (editable, 0, cursor_pos);
    if (!text) return NULL;
    s=text;
    for  (i=0; i<cursor_pos; i++)
        s++;

    *s = '\0';
    while (i>0)
    {
        s--;
        if (*s==' ')
        {
            s++;
            if (*s !='\0' && *s !=' ')
            {
                prefix = g_strdup(s);
                g_free(text);
                return prefix;
            }
            break;
        }
        i--;
    }
    if (!i) prefix = g_strdup(s);
    g_free(text);
    return prefix;
}


static void  entry_dynamic_expansion (GmEntry * entry)
{
    gchar * text_line ;
    g_return_if_fail (entry!=NULL);


    text_line = gtk_entry_get_text (GTK_ENTRY(entry));
    g_return_if_fail (text_line != NULL);
    text_line = g_strdup (text_line);
    g_strstrip(text_line);
    if (strlen(text_line)<1)
    {
        g_free (text_line);
        return;
    }
    entry_cycle_expand(entry);
    return;
}

static void entry_cycle_expand (GmEntry *entry)
{

    GtkEditable * editable;
    static gchar * new_prefix = NULL;
    static gchar * prefix=NULL;
    static GList *compl_items = NULL;
    static GList *item = NULL;
    static gint   cursor_pos=0;
    static gint last_insert_length    = 0;
    gboolean first_time = FALSE;
    gchar * s, *inserted_string = NULL;
    gint i = 0;
    gint new_cursor_pos;
    editable  = GTK_EDITABLE(entry);

    if (!entry->compl_cycling)
    {
        cursor_pos = editable->current_pos;
        if (!(prefix = entry_get_completion_prefix(editable)))
        {
            return;
        }
        entry->compl_cycling = TRUE;
        item = compl_items = g_completion_complete (entry->completion, prefix, &new_prefix);
        first_time = TRUE;
    }
    if (!compl_items)
        return;

    if (!first_time &&  g_list_length (compl_items)==1)
        return;
    if (!item)
        item = compl_items;
    if (strlen(prefix) == strlen (item->data))
        return;

    s = item->data;
    for (i=0;i<strlen(prefix);i++)
        s++;
    if (!first_time)
    {
        /* remove the last inserted completion and restore the cursor position */
        gtk_editable_delete_text (editable, cursor_pos, last_insert_length + cursor_pos);
        editable->current_pos = cursor_pos;
    }
    new_cursor_pos = cursor_pos;
    /* add an empty space */
    inserted_string = g_strdup_printf ("%s",s);
    gtk_editable_insert_text (editable, inserted_string, strlen(inserted_string), &new_cursor_pos);
    gtk_editable_select_region (editable, cursor_pos + (strlen (new_prefix)- strlen(prefix)),new_cursor_pos);
    g_free (inserted_string);
    last_insert_length = new_cursor_pos - cursor_pos;
    editable->current_pos = new_cursor_pos;
    item = item->next;
}


static gint entry_hide_message (gpointer data)
{

    GmEntry * entry = (GmEntry *) data;

    gtk_editable_delete_text (GTK_EDITABLE (entry), 0, -1);

    if (entry->remember_message)
    {
        gtk_entry_set_text (GTK_ENTRY (entry), entry->remember_message);
        GTK_EDITABLE (entry)->current_pos = entry->remember_cursor_pos;
        g_free (entry->remember_message);
        entry->remember_message = NULL;
    }
    gtk_timeout_remove (entry->remove_messages_timer_tag);
    entry->remove_messages_timer_tag   = 0;
    return FALSE;
}


static void entry_show_message (GmEntry * entry, gchar * message)
{
    g_return_if_fail (message != NULL);

    if (entry->remove_messages_timer_tag)
        entry_hide_message (entry);
    entry->remember_message = gtk_editable_get_chars  (GTK_EDITABLE (entry), 0, -1);
    entry->remember_cursor_pos = GTK_EDITABLE (entry)->current_pos;
    gtk_entry_set_text (GTK_ENTRY (entry), message);
    gtk_editable_select_region (GTK_EDITABLE (entry), 0, -1);
    entry->remove_messages_timer_tag  = gtk_timeout_add (REMOVE_MESSAGE_TIMEOUT, entry_hide_message, entry);
}



static void
entry_insert_text (GtkEditable *editable,
                   const gchar *new_text,
                   gint         new_text_length,
                   gint        *position)
{


    (* parent_insert_text)(editable,new_text,new_text_length,position);

}




